@@ -6,6 +6,8 @@
 #define CL_HPP_TARGET_OPENCL_VERSION 300
 #include "qrack/qfactory.hpp"
 
+#define QSIM_CONFIG(numQubits) Qrack::CreateArrangedLayersFull(nw, md, sd, sh, bdt, pg, tn, hy, oc, numQubits, Qrack::ZERO_BCI, nullptr, Qrack::CMPLX_DEFAULT_ARG, false, true, hp)
+
 std::string trim(std::string s)
 {
     // Cut leading, trailing, and extra spaces
@@ -30,11 +32,22 @@ struct QrackObservable {
 
 struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
     bool tapeRecording;
-    bitLenInt allocated_qubits;
+    bool sh;
+    bool tn;
+    bool sd;
+    bool md;
+    bool bdt;
+    bool oc;
+    bool pg;
+    bool hy;
+    bool hp;
+    bool nw;
     size_t shots;
+    Qrack::real1_f noise_param;
     Qrack::QInterfacePtr qsim;
     std::map<QubitIdType, bitLenInt> qubit_map;
     std::vector<QrackObservable> obs_cache;
+    std::vector<Qrack::QInterfaceEngine> simulatorType;
 
     // static constants for RESULT values
     static constexpr bool QRACK_RESULT_TRUE_CONST = true;
@@ -174,8 +187,8 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
             const Qrack::real1 omega = inverse ? -params[0U] : params[2U];
             const Qrack::real1 cos0 = (Qrack::real1)cos(theta / 2);
             const Qrack::real1 sin0 = (Qrack::real1)sin(theta / 2);
-            const Qrack::complex expP = exp(Qrack::I_CMPLX * (phi + omega) / (2 * ONE_R1));
-            const Qrack::complex expM = exp(Qrack::I_CMPLX * (phi - omega) / (2 * ONE_R1));
+            const Qrack::complex expP = exp(Qrack::I_CMPLX * (phi + omega) * HALF_R1);
+            const Qrack::complex expM = exp(Qrack::I_CMPLX * (phi - omega) * HALF_R1);
             const Qrack::complex mtrx[4U]{
                 cos0 / expP, -sin0 * expM,
                 sin0 / expM, cos0 * expP
@@ -208,20 +221,20 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
             }
         }
 
+        QRACK_CONST Qrack::complex NEG_1_CMPLX(-ONE_R1, ZERO_R1);
+        QRACK_CONST Qrack::complex NEG_I_CMPLX(ZERO_R1, ONE_R1);
         QRACK_CONST Qrack::complex SQRT1_2_CMPLX(Qrack::SQRT1_2_R1, ZERO_R1);
         QRACK_CONST Qrack::complex NEG_SQRT1_2_CMPLX(-Qrack::SQRT1_2_R1, ZERO_R1);
         QRACK_CONST Qrack::complex SQRTI_2_CMPLX(ZERO_R1, Qrack::SQRT1_2_R1);
         QRACK_CONST Qrack::complex NEG_SQRTI_2_CMPLX(ZERO_R1, -Qrack::SQRT1_2_R1);
         const Qrack::complex QBRTI_2_CMPLX(ZERO_R1, sqrt(Qrack::SQRT1_2_R1));
         const Qrack::complex NEG_QBRTI_2_CMPLX(ZERO_R1, sqrt(-Qrack::SQRT1_2_R1));
-        QRACK_CONST Qrack::complex ONE_PLUS_I_DIV_2 = Qrack::complex((Qrack::real1)(ONE_R1 / 2), (Qrack::real1)(ONE_R1 / 2));
-        QRACK_CONST Qrack::complex ONE_MINUS_I_DIV_2 = Qrack::complex((Qrack::real1)(ONE_R1 / 2), (Qrack::real1)(-ONE_R1 / 2));
 
         QRACK_CONST Qrack::complex pauliX[4U] = { Qrack::ZERO_CMPLX, Qrack::ONE_CMPLX, Qrack::ONE_CMPLX, Qrack::ZERO_CMPLX };
-        QRACK_CONST Qrack::complex pauliY[4U] = { Qrack::ZERO_CMPLX, -Qrack::I_CMPLX, Qrack::I_CMPLX, Qrack::ZERO_CMPLX };
-        QRACK_CONST Qrack::complex pauliZ[4U] = { Qrack::ONE_CMPLX, Qrack::ZERO_CMPLX, Qrack::ZERO_CMPLX, -Qrack::ONE_CMPLX };
-        QRACK_CONST Qrack::complex sqrtX[4U]{ ONE_PLUS_I_DIV_2, ONE_MINUS_I_DIV_2, ONE_MINUS_I_DIV_2, ONE_PLUS_I_DIV_2 };
-        QRACK_CONST Qrack::complex iSqrtX[4U]{ ONE_MINUS_I_DIV_2, ONE_PLUS_I_DIV_2, ONE_PLUS_I_DIV_2, ONE_MINUS_I_DIV_2 };
+        QRACK_CONST Qrack::complex pauliY[4U] = { Qrack::ZERO_CMPLX, NEG_I_CMPLX, Qrack::I_CMPLX, Qrack::ZERO_CMPLX };
+        QRACK_CONST Qrack::complex pauliZ[4U] = { Qrack::ONE_CMPLX, Qrack::ZERO_CMPLX, Qrack::ZERO_CMPLX, NEG_1_CMPLX };
+        QRACK_CONST Qrack::complex sqrtX[4U]{ Qrack::HALF_I_HALF_CMPLX, Qrack::HALF_NEG_I_HALF_CMPLX, Qrack::HALF_NEG_I_HALF_CMPLX, Qrack::HALF_I_HALF_CMPLX };
+        QRACK_CONST Qrack::complex iSqrtX[4U]{ Qrack::HALF_NEG_I_HALF_CMPLX, Qrack::HALF_I_HALF_CMPLX, Qrack::HALF_I_HALF_CMPLX, Qrack::HALF_NEG_I_HALF_CMPLX };
         QRACK_CONST Qrack::complex hadamard[4U]{ SQRT1_2_CMPLX, SQRT1_2_CMPLX, SQRT1_2_CMPLX, NEG_SQRT1_2_CMPLX };
 
         if ((name == "PauliX") || (name == "CNOT") || (name == "Toffoli") || (name == "MultiControlledX")) {
@@ -336,8 +349,8 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
             const Qrack::real1 omega = inverse ? -params[0U] : params[2U];
             const Qrack::real1 cos0 = (Qrack::real1)cos(theta / 2);
             const Qrack::real1 sin0 = (Qrack::real1)sin(theta / 2);
-            const Qrack::complex expP = exp(Qrack::I_CMPLX * (phi + omega) / (2 * ONE_R1));
-            const Qrack::complex expM = exp(Qrack::I_CMPLX * (phi - omega) / (2 * ONE_R1));
+            const Qrack::complex expP = exp(Qrack::I_CMPLX * (phi + omega) * HALF_R1);
+            const Qrack::complex expM = exp(Qrack::I_CMPLX * (phi - omega) * HALF_R1);
             const Qrack::complex mtrx[4U]{
                 cos0 / expP, -sin0 * expM,
                 sin0 / expM, cos0 * expP
@@ -369,8 +382,18 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
 
     QrackDevice([[maybe_unused]] std::string kwargs = "{}")
         : tapeRecording(false)
-        , allocated_qubits(0U)
+        , sh(true)
+        , tn(true)
+        , sd(true)
+        , md(true)
+        , bdt(false)
+        , oc(true)
+        , pg(false)
+        , hy(false)
+        , hp(false)
+        , nw(false)
         , shots(1U)
+        , noise_param(ZERO_R1_F)
         , qsim(nullptr)
     {
         // Cut leading '{' and trailing '}'
@@ -380,142 +403,66 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
         kwargs = trim(kwargs);
 
         std::map<std::string, int> keyMap;
-        keyMap["'wires'"] = 1;
-        keyMap["'shots'"] = 2;
-        keyMap["'is_hybrid_stabilizer'"] = 3;
-        keyMap["'is_tensor_network'"] = 4;
-        keyMap["'is_schmidt_decomposed'"] = 5;
-        keyMap["'is_schmidt_decomposition_parallel'"] = 6;
-        keyMap["'is_qbdd'"] = 7;
-        keyMap["'is_gpu'"] = 8;
-        keyMap["'is_host_pointer'"] = 9;
-
-        bitLenInt wires = 0U;
-        bool is_hybrid_stabilizer = true;
-        bool is_tensor_network = false;
-        bool is_schmidt_decomposed = true;
-        bool is_schmidt_decomposition_parallel = true;
-        bool is_qbdd = false;
-        bool is_gpu = true;
-        bool is_host_pointer = false;
+        keyMap["'is_hybrid_stabilizer'"] = 1;
+        keyMap["'is_tensor_network'"] = 2;
+        keyMap["'is_schmidt_decomposed'"] = 3;
+        keyMap["'is_schmidt_decomposition_parallel'"] = 4;
+        keyMap["'is_qbdd'"] = 5;
+        keyMap["'is_gpu'"] = 6;
+        keyMap["'is_paged'"] = 7;
+        keyMap["'is_hybrid_cpu_gpu'"] = 8;
+        keyMap["'is_host_pointer'"] =9;
+        keyMap["'noise'"] = 10;
 
         size_t pos;
         while ((pos = kwargs.find(":")) != std::string::npos) {
             std::string key = trim(kwargs.substr(0, pos));
             kwargs.erase(0, pos + 1U);
-
-            if (key == "'wires'") {
-                // Handle if empty
-                // We look for ',' or npos, to respect other Wires value kwargs
-                if (kwargs.find("<Wires = []>") != kwargs.find("<Wires = []>,")) {
-                    continue;
-                }
-
-                // Handle if integer
-                pos = kwargs.find(",");
-                bool isInt = true;
-                for (size_t i = 0; i < pos; ++i) {
-                    if ((kwargs[i] != ' ') && !isdigit(kwargs[i])) {
-                        isInt = false;
-                        break;
-                    }
-                }
-                if (isInt) {
-                    wires = stoi(trim(kwargs.substr(0, pos)));
-                    for (size_t i = 0U; i < wires; ++i) {
-                        qubit_map[i] = i;
-                    }
-                    kwargs.erase(0, pos + 1U);
-
-                    continue;
-                }
-
-                // Handles if Wires object
-                pos = kwargs.find("]>");
-                std::string value = kwargs.substr(0, pos);
-                kwargs.erase(0, pos + 3U);
-                size_t p = value.find("[");
-                value.erase(0, p + 1U);
-                wires = 0U;
-                size_t q;
-                while ((q = value.find(",")) != std::string::npos) {
-                    qubit_map[(QubitIdType)stoi(trim(value.substr(0, q)))] = wires;
-                    ++wires;
-                    value.erase(0, q + 1U);
-                }
-                qubit_map[stoi(trim(value))] = wires;
-                ++wires;
-
-                continue;
-            }
-
             pos = kwargs.find(",");
             std::string value = trim(kwargs.substr(0, pos));
             kwargs.erase(0, pos + 1U);
             const bool val = (value == "True");
             switch (keyMap[key]) {
+                case 1:
+                    sh = val;
+                    break;
                 case 2:
-                    if (value != "None") {
-                        shots = std::stoi(value);
-                    }
+                    tn = val;
                     break;
                 case 3:
-                    is_hybrid_stabilizer = val;
+                    sd = val;
                     break;
                 case 4:
-                    is_tensor_network = val;
+                    md = val;
                     break;
                 case 5:
-                    is_schmidt_decomposed = val;
+                    bdt = val;
                     break;
                 case 6:
-                    is_schmidt_decomposition_parallel = val;
+                    oc =  val;
                     break;
                 case 7:
-                    is_qbdd = val;
+                    pg =  val;
                     break;
                 case 8:
-                    is_gpu =  val;
+                    hy =  val;
                     break;
                 case 9:
-                    is_host_pointer = val;
+                    hp = val;
+                    break;
+                case 10:
+                    noise_param = std::stof(value);
+                    nw = noise_param > ZERO_R1;
                     break;
                 default:
                     break;
             }
         }
 
-        // Construct backwards, then reverse:
-        std::vector<Qrack::QInterfaceEngine> simulatorType;
-
-        if (!is_gpu) {
-            simulatorType.push_back(Qrack::QINTERFACE_CPU);
+        qsim = QSIM_CONFIG(0U);
+        if (noise_param > ZERO_R1) {
+            qsim->SetNoiseParameter(noise_param);
         }
-
-        if (is_qbdd) {
-            simulatorType.push_back(Qrack::QINTERFACE_BDT_HYBRID);
-        }
-
-        if (is_hybrid_stabilizer) {
-            simulatorType.push_back(Qrack::QINTERFACE_STABILIZER_HYBRID);
-        }
-
-        if (is_schmidt_decomposed) {
-            simulatorType.push_back(is_schmidt_decomposition_parallel ? Qrack::QINTERFACE_QUNIT_MULTI : Qrack::QINTERFACE_QUNIT);
-        }
-
-        if (is_tensor_network) {
-            simulatorType.push_back(Qrack::QINTERFACE_TENSOR_NETWORK);
-        }
-
-        // (...then reverse:)
-        std::reverse(simulatorType.begin(), simulatorType.end());
-
-        if (!simulatorType.size()) {
-            simulatorType.push_back(Qrack::QINTERFACE_CPU);
-        }
-
-        qsim = CreateQuantumInterface(simulatorType, wires, Qrack::ZERO_BCI, nullptr, Qrack::CMPLX_DEFAULT_ARG, false, true, is_host_pointer);
     }
 
     QrackDevice &operator=(const QuantumDevice &) = delete;
@@ -524,15 +471,9 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
     QrackDevice &operator=(QuantumDevice &&) = delete;
 
     auto AllocateQubit() -> QubitIdType override {
-        if (allocated_qubits >= qubit_map.size()) {
-            throw std::runtime_error("Catalyst has requested more qubits than exist in device, with "
-                + std::to_string(allocated_qubits) + " allocated qubits. "
-                + "(Set your wires count high enough, for the device.)");
-        }
-        auto it = qubit_map.begin();
-        std::advance(it, allocated_qubits);
-        const QubitIdType label = it->first;
-        ++allocated_qubits;
+        const QubitIdType label = qsim->GetQubitCount();
+        qsim->Allocate(1U);
+        qubit_map[label] = (bitLenInt)label;
         return label;
     }
     auto AllocateQubits(size_t num_qubits) -> std::vector<QubitIdType> override {
@@ -542,8 +483,6 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
         }
         return ids;
     }
-    [[nodiscard]] auto Zero() const -> Result override { return const_cast<Result>(&QRACK_RESULT_FALSE_CONST); }
-    [[nodiscard]] auto One() const -> Result override { return const_cast<Result>(&QRACK_RESULT_TRUE_CONST); }
     auto Observable(ObsId id, const std::vector<std::complex<double>> &matrix,
                     const std::vector<QubitIdType> &wires) -> ObsIdType override
     {
@@ -564,14 +503,25 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
             default:
                 break;
         }
-        obs_cache.push_back(QrackObservable({ basis }, { (bitLenInt)dev_wires[0U] }));
+        std::vector<Qrack::Pauli> bases;
+        bases.reserve(dev_wires.size());
+        for (size_t i = 0U; i < dev_wires.size(); ++i) {
+            bases.emplace_back(basis);
+        }
+        obs_cache.push_back(QrackObservable(bases, dev_wires));
 
         return obs_cache.size() - 1U;
     }
     auto TensorObservable(const std::vector<ObsIdType> &obs) -> ObsIdType override
     {
+        if (!obs.size()) {
+            return -1;
+        }
         QrackObservable o;
         for (const ObsIdType& id : obs) {
+            if (id >= obs_cache.size()) {
+                throw std::invalid_argument("Observable ID not in device cache: " + std::to_string(id));
+            }
             const QrackObservable& i = obs_cache[id];
             o.obs.insert(o.obs.end(), i.obs.begin(), i.obs.end());
             o.wires.insert(o.wires.end(), i.wires.begin(), i.wires.end());
@@ -602,37 +552,28 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
         qsim->M(id);
         // Deallocate
         qsim->Dispose(id, 1U);
-        qubit_map.erase(label);
     }
     void ReleaseAllQubits() override
     {
         // State vector is left empty
-        qsim->Dispose(0U, qsim->GetQubitCount());
-        qubit_map.clear();
+        qsim = QSIM_CONFIG(0U);
+        if (noise_param > ZERO_R1) {
+            qsim->SetNoiseParameter(noise_param);
+        }
     }
     [[nodiscard]] auto GetNumQubits() const -> size_t override
     {
         return qsim->GetQubitCount();
     }
-    void SetDeviceShots(size_t s) override { shots = s; }
+    void SetDeviceShots(size_t s) override {
+        if ((s > 1U) && (noise_param > ZERO_R1_F)) {
+            throw std::domain_error("Shots > 1 can't be simulated with noise on the Qrack back end! (Likely, you want to set mcm_method=\"one-shot\" on your qnode, with multiple shots.)");
+        }
+        shots = s;
+    }
     [[nodiscard]] auto GetDeviceShots() const -> size_t override { return shots; }
     void StartTapeRecording() override { tapeRecording = true; }
     void StopTapeRecording() override { tapeRecording = false; }
-    void PrintState() override
-    {
-        const size_t numQubits = qsim->GetQubitCount();
-        const size_t maxQPower = (size_t)qsim->GetMaxQPower();
-        const size_t maxLcv = maxQPower - 1U;
-        size_t idx = 0U;
-        std::cout << "*** State-Vector of Size " << maxQPower << " ***" << std::endl;
-        std::cout << "[";
-        std::unique_ptr<Qrack::complex> sv(new Qrack::complex[maxQPower]);
-        qsim->GetQuantumState(sv.get());
-        for (; idx < maxLcv; ++idx) {
-            std::cout << sv.get()[idx] << ", ";
-        }
-        std::cout << sv.get()[idx] << "]" << std::endl;
-    }
     void NamedOperation(const std::string &name, const std::vector<double> &params,
                         const std::vector<QubitIdType> &wires, bool inverse,
                         const std::vector<QubitIdType> &controlled_wires,
@@ -710,17 +651,23 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
     }
     auto Expval(ObsIdType id) -> double override
     {
+        if (id >= obs_cache.size()) {
+            throw std::invalid_argument("Observable ID not in device cache: " + std::to_string(id));
+        }
         const QrackObservable& obs = obs_cache[id];
         return qsim->ExpectationPauliAll(obs.wires, obs.obs);
     }
     auto Var(ObsIdType id) -> double override
     {
+        if (id >= obs_cache.size()) {
+            throw std::invalid_argument("Observable ID not in device cache: " + std::to_string(id));
+        }
         const QrackObservable& obs = obs_cache[id];
         return qsim->VariancePauliAll(obs.wires, obs.obs);
     }
     void State(DataView<std::complex<double>, 1>& sv) override
     {
-        RT_FAIL_IF(sv.size() != (size_t)qsim->GetMaxQPower(), "Invalid size for the pre-allocated state vector");
+        RT_FAIL_IF(sv.size() != (size_t)((uint64_t)qsim->GetMaxQPower()), "Invalid size for the pre-allocated state vector");
         reverseWires();
 #if FPPOW == 6
         qsim->GetQuantumState(&(*(sv.begin())));
@@ -733,7 +680,7 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
     }
     void Probs(DataView<double, 1>& p) override
     {
-        RT_FAIL_IF(p.size() != (size_t)qsim->GetMaxQPower(), "Invalid size for the pre-allocated probabilities vector");
+        RT_FAIL_IF(p.size() != (size_t)((uint64_t)qsim->GetMaxQPower()), "Invalid size for the pre-allocated probabilities vector");
         reverseWires();
 #if FPPOW == 6
         qsim->GetProbs(&(*(p.begin())));
@@ -746,7 +693,7 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
     }
     void PartialProbs(DataView<double, 1> &p, const std::vector<QubitIdType> &wires) override
     {
-        RT_FAIL_IF((size_t)Qrack::pow2(wires.size()) != p.size(), "Invalid size for the pre-allocated probabilities vector");
+        RT_FAIL_IF((size_t)((uint64_t)Qrack::pow2(wires.size())) != p.size(), "Invalid size for the pre-allocated probabilities vector");
         auto &&dev_wires = getDeviceWires(wires);
         std::reverse(dev_wires.begin(), dev_wires.end());
 #if FPPOW == 6
@@ -757,108 +704,165 @@ struct QrackDevice final : public Catalyst::Runtime::QuantumDevice {
         std::copy(_p.get(), _p.get() + p.size(), p.begin());
 #endif
     }
-    void Sample(DataView<double, 2> &samples, size_t shots) override
+    void _SampleBody(const size_t numQubits, const std::map<bitCapInt, int>& q_samples, DataView<double, 2> &samples)
     {
-        // TODO: We could suggest, for upstream, that "shots" is a redundant parameter
-        // that could be instead implied by the size of "samples."
-        RT_FAIL_IF(samples.size() != shots, "Invalid size for the pre-allocated samples");
+        auto samplesIter = samples.begin();
+        auto q_samplesIter = q_samples.begin();
+        for (auto q_samplesIter = q_samples.begin(); q_samplesIter != q_samples.end(); ++q_samplesIter) {
+            const bitCapInt sample = q_samplesIter->first;
+            int shots = q_samplesIter->second;
+            for (; shots > 0; --shots) {
+                for (size_t wire = 0U; wire < numQubits; ++wire) {
+                    *(samplesIter++) = bi_compare_0((sample >> wire) & 1U) ? 1.0 : 0.0;
+                }
+            }
+        }
+    }
+    void Sample(DataView<double, 2> &samples) override
+    {
+        RT_FAIL_IF(samples.size() != shots * qsim->GetQubitCount(), "Invalid size for the pre-allocated samples");
+
+        if ((shots > 1U) && (noise_param > ZERO_R1_F)) {
+            throw std::domain_error("Shots > 1 can't be simulated with noise on the Qrack back end! (Likely, you want to set mcm_method=\"one-shot\" on your qnode, with multiple shots.)");
+        }
+
+        if (shots == 1U) {
+            const bitCapInt rev_sample = qsim->MAll();
+            const bitLenInt numQubits = qsim->GetQubitCount();
+            bitCapInt sample = Qrack::ZERO_BCI;
+            for (bitLenInt i = 0U; i < numQubits; ++i) {
+                if (bi_compare_0(rev_sample & Qrack::pow2(i)) != 0) {
+                    sample = sample | Qrack::pow2(numQubits - (i + 1U));
+                }
+            }
+            const std::map<bitCapInt, int> q_samples{{sample, 1}};
+            _SampleBody(numQubits, q_samples, samples);
+            return;
+        }
 
         std::vector<bitCapInt> qPowers(qsim->GetQubitCount());
         for (bitLenInt i = 0U; i < qPowers.size(); ++i) {
-            qPowers[i] = Qrack::pow2(qPowers.size() - (i + 1U));
-        }
-        auto q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
-
-        auto samplesIter = samples.begin();
-        for (size_t shot = 0U; shot < shots; ++shot) {
-            bitCapInt sample = q_samples[shot];
-            for (size_t wire = 0U; wire < qPowers.size(); ++wire) {
-                *(samplesIter++) = bi_to_double((sample >> wire) & 1U);
-            }
+            qPowers[i] = Qrack::pow2(i);
         }
+        const std::map<bitCapInt, int> q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
+        _SampleBody(qPowers.size(), q_samples, samples);
     }
-    void PartialSample(DataView<double, 2> &samples, const std::vector<QubitIdType> &wires, size_t shots) override
+    void PartialSample(DataView<double, 2> &samples, const std::vector<QubitIdType> &wires) override
     {
-        // TODO: We could suggest, for upstream, that "shots" is a redundant parameter
-        // that could be instead implied by the size of "samples."
-        RT_FAIL_IF(samples.size() != shots, "Invalid size for the pre-allocated samples");
+        RT_FAIL_IF(samples.size() != shots * wires.size(), "Invalid size for the pre-allocated samples");
+
+        if ((shots > 1U) && (noise_param > ZERO_R1_F)) {
+            throw std::domain_error("Shots > 1 can't be simulated with noise on the Qrack back end! (Likely, you want to set mcm_method=\"one-shot\" on your qnode, with multiple shots.)");
+        }
 
         auto &&dev_wires = getDeviceWires(wires);
+
+        if (shots == 1U) {
+            const bitCapInt rev_sample = qsim->MAll();
+            const bitLenInt numQubits = dev_wires.size();
+            bitCapInt sample = Qrack::ZERO_BCI;
+            for (bitLenInt i = 0U; i < numQubits; ++i) {
+                if (bi_compare_0(rev_sample & Qrack::pow2(dev_wires[i])) != 0) {
+                    sample = sample | Qrack::pow2(numQubits - (i + 1U));
+                }
+            }
+            const std::map<bitCapInt, int> q_samples{{sample, 1}};
+            _SampleBody(numQubits, q_samples, samples);
+            return;
+        }
+
         std::vector<bitCapInt> qPowers(dev_wires.size());
         for (size_t i = 0U; i < qPowers.size(); ++i) {
-            qPowers[i] = Qrack::pow2(dev_wires[qPowers.size() - (i + 1U)]);
+            qPowers[i] = Qrack::pow2(dev_wires[i]);
         }
-        auto q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
-
-        auto samplesIter = samples.begin();
-        for (size_t shot = 0U; shot < shots; ++shot) {
-            bitCapInt sample = q_samples[shot];
-            for (size_t wire = 0U; wire < qPowers.size(); ++wire) {
-                *(samplesIter++) = bi_to_double((sample >> wire) & 1U);
+        const std::map<bitCapInt, int> q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
+        _SampleBody(qPowers.size(), q_samples, samples);
+    }
+    void _CountsBody(const size_t numQubits, const std::map<bitCapInt, int>& q_samples, DataView<int64_t, 1> &counts)
+    {
+        for (auto q_samplesIter = q_samples.begin(); q_samplesIter != q_samples.end(); ++q_samplesIter) {
+            const bitCapInt sample = q_samplesIter->first;
+            int shots = q_samplesIter->second;
+            std::bitset<1U << QBCAPPOW> basisState;
+            for (size_t wire = 0; wire < numQubits; wire++) {
+                basisState[wire] = bi_compare_0((sample >> wire) & 1U);
+            }
+            for (; shots > 0; --shots) {
+                ++counts(static_cast<size_t>(basisState.to_ulong()));
             }
         }
     }
-    void Counts(DataView<double, 1> &eigvals, DataView<int64_t, 1> &counts,
-                size_t shots) override
+    void Counts(DataView<double, 1> &eigvals, DataView<int64_t, 1> &counts) override
     {
-        // TODO: We could suggest, for upstream, that "shots" is a redundant parameter
-        // that could be instead implied by the size of "eigvals"/"counts".
         const size_t numQubits = qsim->GetQubitCount();
-        const size_t numElements = (size_t)qsim->GetMaxQPower();
+        const size_t numElements = 1U << numQubits;
 
         RT_FAIL_IF(eigvals.size() != numElements || counts.size() != numElements,
                    "Invalid size for the pre-allocated counts");
 
-        std::vector<bitCapInt> qPowers(numQubits);
-        for (bitLenInt i = 0U; i < qPowers.size(); ++i) {
-            qPowers[i] = Qrack::pow2(qPowers.size() - (i + 1U));
+        std::map<bitCapInt, int> q_samples;
+        if (shots == 1U) {
+            const bitCapInt rev_sample = qsim->MAll();
+            const bitLenInt numQubits = qsim->GetQubitCount();
+            bitCapInt sample = Qrack::ZERO_BCI;
+            for (bitLenInt i = 0U; i < numQubits; ++i) {
+                if (bi_compare_0(rev_sample & Qrack::pow2(i)) != 0) {
+                    sample = sample | Qrack::pow2(numQubits - (i + 1U));
+                }
+            }
+            q_samples[sample] = 1;
+        } else {
+            std::vector<bitCapInt> qPowers(numQubits);
+            for (bitLenInt i = 0U; i < qPowers.size(); ++i) {
+                qPowers[i] = Qrack::pow2(i);
+            }
+            q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
         }
-        auto q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
 
         std::iota(eigvals.begin(), eigvals.end(), 0);
         std::fill(counts.begin(), counts.end(), 0);
 
-        for (size_t shot = 0; shot < shots; ++shot) {
-            bitCapInt sample = q_samples[shot];
-            std::bitset<1U << QBCAPPOW> basisState;
-            size_t idx = numQubits;
-            for (size_t wire = 0; wire < numQubits; wire++) {
-                basisState[--idx] = bi_compare_0((sample >> wire) & 1U);
-            }
-            ++counts(static_cast<size_t>(basisState.to_ulong()));
-        }
+        _CountsBody(numQubits, q_samples, counts);
     }
 
     void PartialCounts(DataView<double, 1> &eigvals, DataView<int64_t, 1> &counts,
-                       const std::vector<QubitIdType> &wires, size_t shots) override
+                       const std::vector<QubitIdType> &wires) override
     {
-        // TODO: We could suggest, for upstream, that "shots" is a redundant parameter
-        // that could be instead implied by the size of "eigvals"/"counts".
         const size_t numQubits = wires.size();
-        const size_t numElements = (size_t)Qrack::pow2(numQubits);
+        const size_t numElements = 1U << numQubits;
 
         RT_FAIL_IF(eigvals.size() != numElements || counts.size() != numElements,
                    "Invalid size for the pre-allocated counts");
 
+        if ((shots > 1U) && (noise_param > ZERO_R1_F)) {
+            throw std::domain_error("Shots > 1 can't be simulated with noise on the Qrack back end! (Likely, you want to set mcm_method=\"one-shot\" on your qnode, with multiple shots.)");
+        }
+
         auto &&dev_wires = getDeviceWires(wires);
-        std::vector<bitCapInt> qPowers(dev_wires.size());
-        for (size_t i = 0U; i < qPowers.size(); ++i) {
-            qPowers[i] = Qrack::pow2(dev_wires[qPowers.size() - (i + 1U)]);
+
+        std::map<bitCapInt, int> q_samples;
+        if (shots == 1U) {
+            const bitCapInt rev_sample = qsim->MAll();
+            const bitLenInt numQubits = dev_wires.size();
+            bitCapInt sample = Qrack::ZERO_BCI;
+            for (bitLenInt i = 0U; i < numQubits; ++i) {
+                if (bi_compare_0(rev_sample & Qrack::pow2(dev_wires[i])) != 0) {
+                    sample = sample | Qrack::pow2(numQubits - (i + 1U));
+                }
+            }
+            q_samples[sample] = 1;
+        } else {
+            std::vector<bitCapInt> qPowers(dev_wires.size());
+            for (size_t i = 0U; i < qPowers.size(); ++i) {
+                qPowers[i] = Qrack::pow2(dev_wires[qPowers.size() - (i + 1U)]);
+            }
+            q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
         }
-        auto q_samples = qsim->MultiShotMeasureMask(qPowers, shots);
 
         std::iota(eigvals.begin(), eigvals.end(), 0);
         std::fill(counts.begin(), counts.end(), 0);
 
-        for (size_t shot = 0; shot < shots; ++shot) {
-            bitCapInt sample = q_samples[shot];
-            std::bitset<1U << QBCAPPOW> basisState;
-            size_t idx = numQubits;
-            for (size_t wire = 0; wire < numQubits; wire++) {
-                basisState[--idx] = bi_compare_0((sample >> wire) & 1U);
-            }
-            ++counts(static_cast<size_t>(basisState.to_ulong()));
-        }
+        _CountsBody(numQubits, q_samples, counts);
     }
 
     void Gradient(std::vector<DataView<double, 1>> &, const std::vector<size_t> &) override {}
